##################################################################################
#                                                                                #
#                       RETO MINSAIT REAL ESTATE MODELLING                       #
#                                                                                #
#                              Autores: Data Fever                               #
#                                                                                #
#                            AFI Escuela de Finanzas                             #
##################################################################################


# ------------------------ Carga de paquetes ------------------------ 

import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_absolute_error
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import median_absolute_error
from sklearn.svm import SVR
import xgboost as xgb
from sklearn.model_selection import KFold
from sklearn.ensemble import AdaBoostRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.neighbors import KNeighborsRegressor


# -------------- Definición de funciones auxiliares para el entrenamiento --------------

# Función para entrenar una vez el modelo XGBoost
def xgboost_experiment(df, test, th, param, rate = 1, th_low = 5, th_up = 30):

    # Eliminamos outliers
    train = df[df.TARGET < th]

    # Oversampling
    train = pd.concat([train,augmentate(train[(train.TARGET > th_low) & (train.TARGET < th_up)], rate = rate)])
    
    # División en atributos y variable objetivo
    X_train = train.loc[train.TARGET < th, train.columns!='TARGET']
    y_train = train.loc[train.TARGET < th, 'TARGET']
    
    X_test = test.loc[:, test.columns!='TARGET']
    
    y_max = np.max(train.TARGET)
    
    # Normalizado
    y_train = y_train/y_max
    y_test = np.zeros(len(test))
    
    # Construcción de las matrices de train y test
    dtrain = xgb.DMatrix(X_train, label=y_train)
    dtest = xgb.DMatrix(X_test, label=y_test)

    # Entrenamiento
    param['nthread'] = 4
    param['eval_metric'] = 'mae'
    
    num_round = 40    
    bst = xgb.train(param, dtrain, num_round)
    preds = correct(bst.predict(dtest))
    
    return preds*y_max 


# Función para entrenar n_reg veces XGBoost y promediarlas
def xgb_ensemble(df, test, th, param, rate = 1, th_low = 5, th_up = 30, n_reg = 10):
    preds = pd.DataFrame([])
    
    for i in range(n_reg):
        preds_test = xgboost_experiment(df,test, th, param, rate, th_low, th_up)        
        preds = pd.concat([preds, pd.DataFrame(preds_test)],axis = 1)
        
    ensemble_preds = np.median(preds, axis = 1)  # Ensemble mediante la media

    return ensemble_preds

# Función para predecir a partir de un regresor XGBoost
def eval_xgb(df, bsts, y_max):
    dtest = xgb.DMatrix(df.loc[:, df.columns != 'TARGET'], label = df.TARGET)
    preds = np.array([bst.predict(dtest)*y_max for bst in bsts]).T
    return np.median(preds, axis = 1)

# Función auxiliar para replicar aleatoriamente muestras según la proporción rate
def augmentate(df, rate):
    return df.iloc[np.random.choice(len(df), int(len(df)*(rate)))]

# Función para evitar que haya valores negativos
def correct(preds):
    preds[preds < 0] = 0    
    return preds

# Función para entrenar un modelo realizando k-Fold Cross Validation
def experiment_cv(df, test, reg, th, n_splits = 5, th_low = 5, th_up = 30, rate = 2):
    train = df

    e_in = 0   # Error in-sample
    e_out = 0  # Error out-of-sample
    r_2 = 0    # coeficiente r2
    regs = np.array([]) # Vector de regresores

    # Creación de los conjuntos de validación
    kf = KFold(n_splits = n_splits, random_state = 42, shuffle = True)   
    
    # Bucle de entrenamiento
    for train_index, test_index in kf.split(train):
        train2 = train.iloc[train_index]
        train2 = train2.loc[train2.TARGET < th] # Eliminamos outliers

        # Oversampling y división en atributos y objetivo
        train2 = pd.concat([train2,augmentate(train2[(train2.TARGET > th_low) & (train2.TARGET < th_up)], rate = rate)])
        X = train2.iloc[:,df.columns != 'TARGET']
        y = train2.TARGET    

        test2 = train.iloc[test_index]

        # Ajuste del modelo
        reg = reg.fit(X, y)
        
        # Guardamos valores obtenidos dentro de la muestra
        r_2 += reg.score(X,y) 
        e_in += median_absolute_error(train.TARGET, correct(reg.predict(train.loc[:,train.columns != 'TARGET'])) )
        e_out += median_absolute_error(test2.TARGET, correct(reg.predict(test2.iloc[:, df.columns != 'TARGET'])) )        
        regs = np.append(regs,reg)
    
    # Obtención de predicciones sobre test
    preds = pd.DataFrame([reg.predict(test.iloc[:, df.columns != 'TARGET']) for reg in regs]) 
    preds = np.median(preds, axis = 0)
    
    print("Valores en validación: ")    
    print([r_2/n_splits, e_in/n_splits, e_out/n_splits])
    return preds


# -------------------- Construcción del modelo final --------------------


# Lectura de los archivos de train y test
df = pd.read_csv("train.csv", sep = "|")
test = pd.read_csv("test.csv", sep = "|")


# Columnas a utilizar en el entrenamiento
cols = ['HY_metros_totales','GA_mean_bounce', 'GA_exit_rate', 'GA_page_views','dif','NUM_FOTOS','len_description','log_views','HY_precio','HY_antiguedad','NUM_PESO_MEDIO','TARGET']

# DataFrame para almacenar las predicciones de cada regresor sobre el conjunto test
tags_ensemble = pd.DataFrame([])

# Obtenemos cada uno de los modelos simples para el ensemble

# Entrenamiento y obtención de predicciones de AdaBoost
reg = AdaBoostRegressor(DecisionTreeRegressor(max_depth = 12),learning_rate = 0.003, n_estimators = 15, random_state=42)
preds = experiment_cv(df.loc[:,cols],test.loc[:,cols],reg,95, th_low = 5, th_up = 29, rate = 2.8)
tags = pd.Series(correct(preds))
tags_ensemble = pd.concat([tags_ensemble,tags], axis = 1)

# Entrenamiento y obtención de predicciones de RandomForest
reg = RandomForestRegressor(max_depth=30, random_state=0, n_estimators=90,min_samples_split = 80)
preds = experiment_cv(df.loc[:,cols],test.loc[:,cols],reg,90, th_low = 5, th_up = 29, rate = 2.2)
tags = pd.Series(correct(preds))
tags_ensemble = pd.concat([tags_ensemble,tags], axis = 1)

# Entrenamiento y obtención de predicciones de XGBoost
param = {'max_depth': 3, 'eta': 0.63, 'silent': 1, 'objective': 'reg:linear', 'lambda':0.5, 'alpha':0.0008, 'rate_drop':0.95}
preds  = xgb_ensemble(df.loc[:,cols], test.loc[:,cols], 72, param)
tags = pd.Series(correct(preds))
tags_ensemble = pd.concat([tags_ensemble,tags], axis = 1)

tags_ensemble.columns = ['AdaBoost', 'RF','XGB']


# Leemos los modelos obtenidos para todas las variables numéricas
p = pd.read_csv('predictions.csv', sep = ',', encoding = 'utf-8')
tags_ensemble['RF2'] = p['RF']
tags_ensemble['GB'] = p['GB']


# Parámetros para el ensemble. El mejor modelo es el que tiene más peso,
# mientras que el resto tienen igual influencia
alpha = 0.12
beta = 0.12
gamma = 0.12
psi = 0.12


# Obtención de las predicciones finales mediante el ensemble en el conjunto de test
test['TARGET'] = alpha*tags_ensemble.AdaBoost + beta * tags_ensemble.RF + gamma * tags_ensemble.XGB + \
                     psi*tags_ensemble.RF2 + (1 - alpha - beta - gamma - psi) * tags_ensemble.GB
test.loc[test.TARGET < 15,'TARGET'] = 15
test[['HY_id', 'TARGET']].to_csv('./DataFever.txt', sep = '|', encoding = 'utf-8', index = False)